/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2007, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Bridging API Test Application
 *
 * \author Joshua Colp <jcolp@digium.com>
 * 
 * This is a test application for the new bridging API.
 * \ingroup applications
 */

#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/app.h"
#include "asterisk/dial.h"
#include "asterisk/bridging.h"

static char *app = "BridgeTest";
static char *synopsis = 
"Bridging API Test Application.";
static char *descrip = "Bridging API Test Application.\n";

static int bridge_test_exec(struct ast_channel *chan, void *data)
{
	struct ast_dial *dial = NULL;
	struct ast_bridge *bridge0 = NULL, *bridge1 = NULL;
	char *tech = NULL, *resource = NULL;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "%s requires an argument (Tech/Resource)\n", app);
		return -1;
	}

	/* Create local duplicate for manipulation */
	tech = ast_strdupa(data);

	/* Split up technology and resource */
	if (!(resource = strchr(tech, '/'))) {
		ast_log(LOG_WARNING, "%s requires an argument (Tech/Resource)\n", app);
		return -1;
	}
	*resource++ = '\0';

	/* Okay... create a new dialing structure */
	if (!(dial = ast_dial_create()))
		return -1;

	ast_dial_append(dial, tech, resource);

	/* Actually dial what we want to call */
	if (ast_dial_run(dial, chan, 0) != AST_DIAL_RESULT_ANSWERED || !ast_dial_answered(dial)) {
		ast_log(LOG_WARNING, "Failed to call %s\n", (char*)data);
		ast_dial_destroy(dial);
		return -1;
	}

	/* Create a new bridge to put the channel that answered into */
	bridge0 = ast_bridge_new(AST_BRIDGE_CAPABILITY_1TO1MIX, AST_BRIDGE_FLAG_DISSOLVE | AST_BRIDGE_FLAG_SMART);

	/* Impart it onto the bridge... */
	ast_bridge_impart(bridge0, ast_dial_answered(dial), NULL, NULL);

	/* Create a new bridge to put the channel that called into */
	bridge1 = ast_bridge_new(AST_BRIDGE_CAPABILITY_1TO1MIX, AST_BRIDGE_FLAG_DISSOLVE | AST_BRIDGE_FLAG_SMART);

	/* Impart this one as well onto the bridge... */
	ast_bridge_impart(bridge1, chan, NULL, NULL);

	/* Merge them together */
	ast_bridge_merge(bridge0, bridge1);

	/* Destroy the second bridge since it's over */
	ast_bridge_destroy(bridge1);

	/* Hang out for a bit */
	usleep(10000000);

	return 0;
}

static int unload_module(void)
{
	return ast_unregister_application(app);
}

static int load_module(void)
{
	return ast_register_application(app, bridge_test_exec, synopsis, descrip);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Bridging API Test Application");
